static const char *MARPAESLIFLUA_NEWKEYWORDS =
  "\n"
  "do -- base function reimplementations\n"
  "   do _G.rawtype = _G.rawtype or _G.type\n"
  "      local meta, override\n"
  "      function _G.type(v)\n"
  "         local vtype = rawtype(v)\n"
  "         if vtype ~= 'table' and vtype ~= 'userdata' then return vtype end\n"
  "         meta = debug.getmetatable(v)\n"
  "         if meta then\n"
  "            override = rawget(meta, '__type')\n"
  "            if override then\n"
  "               if rawtype(override) == 'function' then\n"
  "                  return override(v) or rawtype(v)\n"
  "               else\n"
  "                  return override\n"
  "               end\n"
  "            end\n"
  "         end\n"
  "         return rawtype(v)\n"
  "      end\n"
  "   end\n"
  "   do _G.rawnext = _G.rawnext or _G.next\n"
  "      local meta, override\n"
  "      function _G.next(v, k)\n"
  "         meta = debug.getmetatable(v)\n"
  "         if meta then\n"
  "            override = rawget(meta, '__next')\n"
  "            if override then\n"
  "               if rawtype(override) == 'function' then\n"
  "                  return override(v, k)\n"
  "               else\n"
  "                  return override\n"
  "               end\n"
  "            end\n"
  "         end\n"
  "         return rawnext(v, k)\n"
  "      end\n"
  "   end\n"
  "   do _G.rawexists = _G.rawexists or _G.exists\n"
  "      local meta, override\n"
  "      function _G.exists(v, k)\n"
  "         meta = debug.getmetatable(v)\n"
  "         if meta then\n"
  "            override = rawget(meta, '__exists')\n"
  "            if override then\n"
  "               if rawtype(override) == 'function' then\n"
  "                  return override(v, k)\n"
  "               else\n"
  "                  return override\n"
  "               end\n"
  "            end\n"
  "         end\n"
  "         return rawexists(v, k)\n"
  "      end\n"
  "   end\n"
  "   do _G.rawdelete = _G.rawdelete or _G.delete\n"
  "      local meta, override\n"
  "      function _G.delete(v, k)\n"
  "         meta = debug.getmetatable(v)\n"
  "         if meta then\n"
  "            override = rawget(meta, '__delete')\n"
  "            if override then\n"
  "               if rawtype(override) == 'function' then\n"
  "                  return override(v, k)\n"
  "               else\n"
  "                  return override\n"
  "               end\n"
  "            end\n"
  "         end\n"
  "         return rawdelete(v, k)\n"
  "      end\n"
  "   end\n"
  "end\n"
  "\n";
