use strict;
use warnings;
use Test::More;

# generated by Dist::Zilla::Plugin::Test::PodSpelling 2.007005
use Test::Spelling 0.12;
use Pod::Wordlist;


add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( bin lib ) );
__DATA__
BNF
Bindings
Calculator
Durand
ESLIF
Event
EventBitSet
Grammar
Interface
Introduction
Jean
Level
Logger
MarpaX
Properties
PropertyBitSet
Recognizer
Rule
String
Symbol
Tutorial
Type
Value
jeandamiendurand
lib
